//////////////////////////////////////
//
// Reference Calculations and Measurements
//
//////////////////////////////////////
//
// DAC
// 12 bit DAC, 2.048V reference
// .0005V per count
//
// HV264 Amp
// Gain = 66.7
// .03335V per count
//
// .03335V per step
// 136.6V Max out (134)
// 2.048V max input at chip with 1X ref gain
// 144V HV PS to chip
// Measured 100TH Bias Vg 39.45V 1182 0x049E
// Measured 6BN11 Vs 101.5V 3043 0x0BE3
//
//////////////////////////////////////
//
// Power Tube Current Sense
// 
// High Side Current Sense Board (TLP7920 Chip)
// 10 bit ADC, 0.55V ref
// A = 4.1, offset 1.25V
// 1.0490417 Ohm Sense
// 480 cnts for 120mA (4 counts/mA, 0.002148V/mA)
// 120mA -> .125885V -> .516129V -> .258065V
//
// Conventional cathode sense should use 2.148 Ohm sense resistor
//
// Current code subtracts out the current from feedback
//   resistor in the Corona Amp. Code needs to be adjusted
//   for conventional cathode current sense
//
//////////////////////////////////////
//
// HV/Plate Voltage Sense
//
// I used a sense divider of 10M -> 5379R
// Then I RC filtered with 100k/3.3uF cap
// Output of RC filter connected to ADC pin
//
//////////////////////////////////////
//
// Bias Screen Gain Measurements 6BN11
//
// 102.1V -> 240V
// 103.1V -> 195V
// A = ~-50
//
//////////////////////////////////////
//
// Bias Gain Measurements 100TH
//
// 36.47V -> 117.5mA
// 34.47V -> 110.5mA
// A = ~3.5mA/V
//
//////////////////////////////////////

#ifndef _AMP_CONTROL_H
#define _AMP_CONTROL_H

//////////////////////////////////////
//
// Conditional compilation defines
//
//////////////////////////////////////

// Output Tube Choice, only define one at a time
#define OUTPUT_TUBE_100TH
//#define OUTPUT_TUBE_327A
//#define OUTPUT_TUBE_VT_127A
//#define OUTPUT_TUBE_211

//#define AMP_CONTROL_DEBUG // serial port outputs for debugging
//#define HOURS_METER_INITIALIZE // enables code that resets hours meter to zero in setup()
//#define BIAS_DEFAULTS_INITIALIZE //enables code that sets default bias values in EEPROM in setup()

//////////////////////////////////////
//
// Configurables
//   ALL TIMES IN MILLISECONDS
//
//////////////////////////////////////

// tube warmup times
const unsigned long INPUT_TUBE_WARMUP_TIME = 35000;
const unsigned long OUTPUT_TUBE_WARMUP_TIME = 15000;
// time to wait before low HV fault will be monitored
const unsigned long HV_SETTLE_TIME = 5000;

// amplifier will shut off after timeout, set to 0 for no timeout
// 600000 for 10 minutes
const unsigned long SILENCE_TIMEOUT = 600000;

// Silence threshold used for automatic shutdown
const unsigned int SILENCE_THRESHOLD = 186; //100mV

// faults must persist longer than this number
// sometimes EMI causes false alarms, so this ensures fault is real
const unsigned long FAULT_PERSIST_LIMIT = 200;

// time the bias ramp will take
const long BIAS_RAMP_TIME = 5000;

// must be larger than the adjust time, lets the bias settle before next adjustment
const long BIAS_TOTAL_TIME = 20000;

// minimum display update period in milliseconds, makes display less jittery
const unsigned long MIN_DISPLAY_UPDATE_PERIOD = 200; 

// Autobias information for 6BN11
const long IT_BIAS_VOLTAGE_TARGET = 320;
const long IT_PLATE_NEAR_SATURATION_VOLTAGE = 50;
const long IT_PLATE_NEAR_HV_VOLTAGE = 400;
const float IT_SLOW_BIAS_GAIN = 0.25;
const float IT_FAST_BIAS_GAIN = 0.33;

// 6BN11 Autobias hard limits, 93.5V nominal screen voltage
const unsigned long IT_BIAS_DAC_MAX = 0x00000EA4; // 125V
const unsigned long IT_BIAS_DAC_MIN = 0x000005DB; // 50V

// EEPROM default values for writing during EEPROM 
// initialization (only when compile flag is set)
// .033V/count with HV amp chip
#define EEPROM_IT1_BIAS_VALUE 0x00000BE3 // 100.6V
#define EEPROM_IT2_BIAS_VALUE 0x00000BE3 // 99.6V

// Autobias information for output tubes
#if defined (OUTPUT_TUBE_211)
  const bool PT_BIAS = false; // 211 not under uC control
  const unsigned long HV_FAULT_VALUE = 700; // in V, nominal 615
  const unsigned long PT_FAULT_VALUE = 75; // output tube rating in W
  const long PT_BIAS_CURRENT_TARGET = 110; // in mA
  const long PT_BIAS_POWER_LIMIT = 72; // in W

// power tube info for 100TH
#elif defined (OUTPUT_TUBE_100TH)
  const bool PT_BIAS = true; // 100TH under uC control
  const unsigned long HV_HIGH_FAULT_VALUE = 825; // in V, nominal 750
  const unsigned long HV_LOW_FAULT_VALUE = 675; // in V, nominal 750
  const unsigned long PT_FAULT_VALUE = 100; // output tube rating in W
  const long PT_BIAS_CURRENT_TARGET = 120; // in mA TO DO: set to 125 after testing
  const long PT_BIAS_POWER_LIMIT = 93; // in W

  // Autobias gain information for 100TH
  const float PT_SLOW_BIAS_GAIN = 1;
  const float PT_FAST_BIAS_GAIN = 2;

  // Autobias hard limits
  #define PT_BIAS_DAC_MAX 0x000005EB // 50V
  #define PT_BIAS_DAC_MIN 0x00000096 // 5V

  // EEPROM default values for writing during EEPROM 
  // initialization (only when compile flag is set)
  // .033V/count with HV amp chip
  #define EEPROM_PT1_BIAS_VALUE 0x000002EE // 25V
  #define EEPROM_PT2_BIAS_VALUE 0x000002EE // 25V
#endif

//////////////////////////////////////
//
// Fixed Design Values
//
//////////////////////////////////////

// Hours Meter defines
const unsigned long DISPLAY_HOURS_METER = 0;
#define EEPROM_HOURS_METER_DEFAULT_VALUE 0x00000000
#define EEPROM_HOURS_METER_ADDRESS 144

// EEPROM addresses for default value storage
#define EEPROM_PT1_BIAS_ADDRESS 128
#define EEPROM_PT2_BIAS_ADDRESS 132
#define EEPROM_IT1_BIAS_ADDRESS 136
#define EEPROM_IT2_BIAS_ADDRESS 140

// Digital Input Pins
#define POWER_BUTTON 2
#define POWER_BUTTON_PRESSED LOW
#define BIAS_BUTTON 3
#define BIAS_BUTTON_PRESSED LOW
#define BACKLIGHT_SWITCH 4
#define BACKLIGHT_SWITCH_ON HIGH

// Analog input pins
#define CHANNEL_A_DRIVER_SENSE_PIN A0
#define CHANNEL_B_DRIVER_SENSE_PIN A1
#define CHANNEL_A_FINAL_SENSE_PIN A2
#define CHANNEL_B_FINAL_SENSE_PIN A3
#define SILENCE_SENSE_PIN A6
#define HV_SENSE_PIN A7

// Digital Output Pins
#define POWER_SSR 10

// Analog Outputs
#define PT1_DAC_OUTPUT MCP4728_CHANNEL_A // power tube 1
#define PT2_DAC_OUTPUT MCP4728_CHANNEL_B // power tube 2
#define IT1_DAC_OUTPUT MCP4728_CHANNEL_C // input tube 1
#define IT2_DAC_OUTPUT MCP4728_CHANNEL_D // input tube 2

// DACs
const uint8_t DAC_ADDRESS = 0x64;

// LCD
const uint8_t LCD_ADDRESS = 0;

//////////////////////////////////////
//
// Type/function definitions
//
//////////////////////////////////////

// amplifier fault codes
enum amp_fault { FAULT_NONE = 0, FAULT_HV_HIGH, FAULT_HV_LOW, FAULT_PT1_OVERPOWER, FAULT_PT2_OVERPOWER, FAULT_IT1_BIAS_HIGH, FAULT_IT2_BIAS_HIGH,
                 FAULT_IT1_BIAS_LOW, FAULT_IT2_BIAS_LOW, FAULT_OT1_BIAS_HIGH, FAULT_OT2_BIAS_HIGH, FAULT_OT1_BIAS_LOW,
                 FAULT_OT2_BIAS_LOW, FAULT_BIAS_ARITHMETIC_PT, FAULT_BIAS_ARITHMETIC_IT, FAULT_BIAS_ADJUST, SILENCE_SHUTDN};

// amplifier state variable type
enum amp_state { AMPLIFIER_OFF, AMPLIFIER_WARMUP, AMPLIFIER_ON };

// bias process state variable type
enum bias_state_type { BIAS_IDLE = 0, BIAS_ADJUST };

// main state task functions
void amplifier_off_task();
void amplifier_warmup_task();
void amplifier_on_task();

// state transition functions
void setup();
void transition_to_amplifier_off(amp_fault fault_code);

// state utility functions
void bias_subtask();
void read_ADCs_and_update_LCD(unsigned long amp_start_time);
int check_for_faults();
void update_builtinLED();

// minor utility functions
void print_by_digit(unsigned int print_value, int num_digits);
unsigned int decimal_place_divisor (int num_digits);

#endif